from datetime import datetime, timedelta
from boto3 import resource as boto3_resource
from functools import cache
from json import load
from os.path import expanduser
from roguewave.modeldata.timebase import ModelTimeConfiguration, \
    timebase_forecast, timebase_lead, timebase_evaluation
from typing import List, Tuple, Dict


# Model private variables
# =============================================================================

# Load the model configuration file.
try:
    with open(expanduser('~/model_configuration.json'), 'rb') as file:
        MODELS = load(file)
except FileNotFoundError as e:
    MODELS = {}


# Classes
# =============================================================================


class _RemoteResourceSpecification:
    """
    Class containing the key template and other necessary information to
    reconstruct uri's on the remote resource
    """

    def __init__(self,
                 name: str,
                 uri_path_template: str,
                 filetype: str = 'netcdf',
                 scheme: str = 's3',
                 mapping_variable_name_model_to_sofar: Dict[str,str] = None,
                 single_variable_per_file: bool = True,
                 model_time_configuration: dict = None):

        """
        :param name: Model name
        :param bucket: bucket
        :param uri_path_template: template of the AWS key
        :param filetype: filetype of the data
        :param model_time_configuration: Description on the timebase of the
        model (how often is the mode run, at what interval is output etc.).
        """
        self.name = name
        self.scheme = scheme
        self.uri_path_template = uri_path_template
        self.filetype = filetype
        self.mapping_variable_name_model_to_sofar = \
            mapping_variable_name_model_to_sofar
        self.single_variable_per_file = single_variable_per_file

        if model_time_configuration is None:
            self.model_time_configuration = ModelTimeConfiguration()
        else:
            self.model_time_configuration = ModelTimeConfiguration(
                **model_time_configuration
            )

    def to_sofar_variable_name(self,variable_name: str) -> str:
        """
        map model variable name to sofar variable name
        :param variable_name: model or sofar variable name
        :return: equivalent sofar_variable_name
        """
        if self.mapping_variable_name_model_to_sofar is not None:
            if variable_name in self.mapping_variable_name_model_to_sofar:
                return self.mapping_variable_name_model_to_sofar[
                    variable_name]
            else:
                # if not in the forward mapping, this may already be a sofar
                # variable name, see if this maps back to a model name, if so
                # this is a variable available in this model.
                try:
                    _ = self.to_model_variable_name(variable_name)
                    return variable_name

                except KeyError as e:
                    raise KeyError( f" Modem {self.name} does not have "
                                    f" output that corresponds to a variable "
                                    f" with name {variable_name}" )
        else:
            # if no variable definitions are available, we assume that this is
            # a sofar model, and the mapping is 1 to 1 i.e.
            #       model_variable_name = sofar_variable_name
            return variable_name

    def to_model_variable_name(self,variable_name: str) -> str:
        """
        map sofar variable name to model variable name
        :param variable_name: sofar or model variable name
        :return: equivalent model_variable_name
        """
        if self.mapping_variable_name_model_to_sofar is not None:
            for key,item in self.mapping_variable_name_model_to_sofar.items():
                if item==variable_name:
                    return key
                elif key==variable_name:
                    return key
            else:
                raise KeyError(f'no mapping for variable {variable_name} for model '
                               f'{self.name}')
        else:
            # if no variable definitions are available, we assume that this is
            # a sofar model, and the mapping is 1 to 1 i.e.
            #       model_variable_name = sofar_variable_name
            return variable_name


# Main Functions
# =============================================================================


def available_models() -> List[str]:
    """
    List all available models
    :return: Available model names as list
    """
    return list(MODELS.keys())


def model_timebase_forecast(
        model:str, init_time: datetime, duration: timedelta
        ) -> List[Tuple[datetime, timedelta]]:
    """
    Get all the valid times for a given forecast up to the requested duration.
    Valid times are returned as a List of pairs: (init_time, forecast_hour)

    :param model: name of the model to retrieve
    :param init_time: init time of the forecast
    :param duration: Duration requested
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """

    aws_layout = _get_resource_specification(model)
    return timebase_forecast(
        init_time=init_time,
        duration=duration,
        time_configuration=aws_layout.model_time_configuration
    )


# -----------------------------------------------------------------------------


def model_timebase_lead(
                  model:str,
                  start_time: datetime,
                  end_time: datetime,
                  lead_time: timedelta,
                  exact:bool = True
                  ) -> List[Tuple[datetime, timedelta]]:
    """
    Get all the valid times at a given lead-time in the period of interest.
    Valid times are returned as a List of pairs: (init_time, forecast_hour).
    If exact is true, only valid-times at exactly the requested lead-time are
    returned. Otherwise, it includes close valid_times as well.

    :param model: name of the model to retrieve
    :param start_time: start time of interval of interest
    :param end_time: end time of interval of interest
    :param lead_time:  lead time of interest
    :param exact: only return valid times that exactly match the requested lead
        time.
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """
    aws_layout = _get_resource_specification(model)
    return timebase_lead(
        start_time=start_time,
        end_time=end_time,
        lead_time=lead_time,
        time_configuration=aws_layout.model_time_configuration,
        exact=exact)


# -----------------------------------------------------------------------------


def model_timebase_evaluation(model:str,
                        evaluation_time: datetime,
                        maximum_lead_time: timedelta,
                        ) -> List[Tuple[datetime, timedelta]]:
    """
    Find all init times and forecast hours so that
        init_time + forecast = evaluation_time,

    :param model: name of the model to retrieve
    :param evaluation_time: Time we want to compare different forecasts for
    :param maximum_lead_time: Maximum lead time of the forecasts to consider
    :return: List of (init_time, forecast_hour) so that evaluation_time =
        init_time + forecast_hour
    """
    aws_layout = _get_resource_specification(model)
    return timebase_evaluation(
        evaluation_time=evaluation_time,
        maximum_lead_time=maximum_lead_time,
        time_configuration=aws_layout.model_time_configuration
    )


# -----------------------------------------------------------------------------
def _generate_aws_key(variable, init_time: datetime, forecast_hour: timedelta,
                      model: _RemoteResourceSpecification) -> str:
    """
    Function that generates a valid AWS key for the given model that represents
    date from the model initialized at the given init_time and with the given
    forecast_hour or lead time.

    :param variable: variable name we want to retrieve (as used in key, if
        applicable, set to None otherwise)
    :param init_time: init time of the model we want to retrieve
    :param forecast_hour: the forecast hour of the model we want to retrieve
    :param model: Model() information on where data is stored.
    :return: A valid aws key for the given model.
    """

    # Helper functions. These are local for now and tied to this function
    # alone. Hence the encapsulation.
    def _replace(_string: str, key: str, value: str):
        """
        Replace a key in string with value
        :param _string: target
        :param key:  string to replace
        :param value:  string to replace with
        :return: string
        """
        return _string.replace("{" + key + "}", value)

    #

    def _time(_string, key, value: datetime):
        """
        Replace {key:fmt} with value.strftime(fmt)
        :param _string: target
        :param key:  either "init_time" or "valid_time"
        :param value: init_time or valid_time as datetime
        :return: string
        """

        # There may be multiple instances that need to be replaced:
        while key in _string:
            # Get the time format string
            _format = _find_between(_string, '{' + key + ':', '}')
            # get the text we want to replace
            to_replace = key + ':' + _format
            # what we want to replace it with (the formatted time)
            replace_with = value.strftime(_format)
            # replace
            _string = _replace(_string, to_replace, replace_with)
        return _string

    #

    def _ecmwf_special(_string, key, _init_time: datetime,
                       _forecast_hour: timedelta):
        """
        Special parsing only valid for ecmwf formatted output.
        :param _string: string
        :param key:  ecmwf_subcycle_string or ecmwf_subcycle_string
        :param _init_time: init time as datetime
        :param _forecast_hour: forecast hour as timedelta
        :return:
        """

        if key == 'ecmwf_lead_zero_indicator':
            # the suffix of ecmwf forecasts is either `001` for all forecast
            # hours exept the zero hour.
            if _forecast_hour == timedelta(hours=0):
                ecmwf_lead_zero_indicator = '011'
            else:
                ecmwf_lead_zero_indicator = '001'
            return _replace(_string, key, ecmwf_lead_zero_indicator)

        elif key == 'ecmwf_subcycle_string':
            # the prefix of ecmwf forecasts is A2P for 10 day forecasts (00Z
            # and # 12Z cycles), or A2Q for the 4 day forecasts (06Z and 18Z
            # cycles)
            if (_init_time.hour == 0) or (_init_time.hour == 12):
                ecmwf_subcycle_string = 'A2P'
            else:
                ecmwf_subcycle_string = 'A2Q'
            return _replace(_string, key, ecmwf_subcycle_string)
        else:
            raise Exception(f'Unknown key {key}')

    # ---------------------

    #
    # Purpose:
    #
    #   The target template key has the form
    #   "t/{init_time:%Y%m%d}/{init_time:%H}/t/{forecasthour}.{variable}.nc"
    #   here everything between {...} defines a token that needs to be replaced
    #   at executing time by a value. E.g. {init_time:%Y%m%d} denotes we want
    #   to replace this with the init_time formated according to %Y%m%d. This
    #   function does the parsing/replacement.
    #
    # Implementation:
    #
    #  We define a set of properties below, where the key denotes the keyword
    #  we want to replace (init_time, forecast_hour etc.), and the value is
    #  a dictionary with a method denoting the function handling the
    #  replacement and a set of arguments in addition to the string and the key
    #  to be replaced.
    valid_time = init_time + forecast_hour

    # Define keywords
    keywords = {
        "forecasthour": {
            'method': _replace, 'args': (
                "{hours:03d}".format(
                    hours=int(forecast_hour.total_seconds() // 3600)
                ),
            )
        },
        "init_time": {'method': _time, 'args': (init_time,)},
        "valid_time": {'method': _time, 'args': (valid_time,)},
        "variable": {'method': _replace, 'args': (variable,)},
        "ecmwf_subcycle_string": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        },
        "ecmwf_lead_zero_indicator": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        }
    }
    # Loop over properties and invoke method to replace keywords.
    uri = model.uri_path_template
    for keyword, action in keywords.items():
        if keyword in uri:
            uri = action['method'](uri, keyword, *action['args'])

    return uri


# -----------------------------------------------------------------------------


@cache
def list_available_variables(model_name, init_time: datetime = None):
    """
    List available variables from a given model. This function _only_ works
    if the variable names are encoded in the aws key. For e.g. ECMWF this
    currently does _not_ work.

    Basically, we look on AWS for the pattern given in model.key_template and
    find all keys conforming to the template with different values for the
    variable name. May be very slow if the prefix (everything before the
    variable name in the template) does not sufficiently restrict the search
    space.

    Note response is cached so that multiple calls are fast.

    :param model_name: the name of the model
    :param init_time: init_time, an init_time for which results from the model
    are available. Defaults to 2022-06-01; this may not work for all models.

    :return: list of variables for the model.
    """

    init_time = datetime(2022, 6, 1) if init_time is None else init_time
    # get model aws key template description
    model = _get_resource_specification(model_name)

    if model.mapping_variable_name_model_to_sofar is not None:
        return list(model.mapping_variable_name_model_to_sofar.keys())

    # open s3 resource
    s3 = boto3_resource('s3')
    my_bucket = s3.Bucket(model.bucket)

    if 'variable' not in model.uri_path_template:
        raise ValueError('Listing variables for a particular model only works'
                         'if the variable name is part of the aws keys.')

    # replace "variable" with "####" as a seperator between everything before
    # and after the variable name in the aws key.
    key = _generate_aws_key(
        '####', init_time, forecast_hour=timedelta(hours=0), model=model)

    # Split in prefix and suffix, and remove the bucket name from the prefix
    prefix, suffix = key.split('####')
    prefix = prefix.replace(model.bucket + '/', '')

    # loop over all objects conforming to the prefix
    variables = []
    for obj in my_bucket.objects.filter(Prefix=prefix):
        if suffix in obj.key:
            # If the suffix is part of the key, add the part in between the
            # prefix and the suffix of the aws key as a variable name to the
            # output list.
            variables.append(
                _find_between(obj.key, prefix, suffix)
            )

    return variables


# Helper functions
# =============================================================================


def _get_resource_specification(name) -> _RemoteResourceSpecification:
    """
    Return the object representing key layout on aws
    :param name:
    :return: Model
    """
    return _RemoteResourceSpecification(name=name, **MODELS[name])


# -----------------------------------------------------------------------------


def _find_between(s: str, first: str, last: str) -> str:
    """
     Simple function that returns the string between the "first" and "last"
     token. E.g. if first='{' and last=']' with string 'asdd{THIS]asdsd'
     this returns 'THIS'.
    :param s: string
    :param first: token signifying start
    :param last:  token signifying end
    :return:  string
    """

    # lets avoid regex for now :-)
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""



from .wavefields import bulk_parameters_partitions, DEFAULT_CONFIG_PARTITION_SPECTRA
from roguewave.wavespectra import WaveSpectrum1D
from roguewave.wavespectra.operators import spectrum1D_time_filter
from roguewave.wavespectra.estimators import spec2d_from_spec1d
from .wavefields import partition_spectra
from typing import List, Dict, Union
from datetime import timedelta
from pandas import DataFrame
from roguewave.wavespectra import WaveSpectrum2D
from typing import overload
from roguewave import logger
from roguewave.wavespectra import logger

default_config = {
    'smoothInTime': False,
    'estimator': {
        'method': 'mem2',
        'numberOfDirections': 36,
        'frequencySmoothing': False,
        'smoothingLengthscale': 1
    },
    'partition_spectra':DEFAULT_CONFIG_PARTITION_SPECTRA
}


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
@overload
def get_spectral_partitions_from_observations(
        spectra: Dict[str, List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None) -> Dict[str, List[List[WaveSpectrum2D]]]: ...


@overload
def get_spectral_partitions_from_observations(
        spectra: List[WaveSpectrum1D],
        minimum_duration: timedelta,
        config=None) -> List[List[WaveSpectrum2D]]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: Dict[str, List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Dict[str, List[DataFrame]]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: List[WaveSpectrum1D],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> List[DataFrame]: ...


# -----------------------------------------------------------------------------
#                       Implementation
# -----------------------------------------------------------------------------


def get_bulk_partitions_from_observations(
        spectra: Union[Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Union[Dict[str, List[DataFrame]], List[DataFrame]]:
    #
    spectral_partitions = get_spectral_partitions_from_observations(
        spectra, minimum_duration, config, verbose)

    if isinstance(spectra, dict):
        output = {}
        for key in spectral_partitions:
            output[key] = bulk_parameters_partitions(spectral_partitions[key])

    elif isinstance(spectra, list):
        output = bulk_parameters_partitions(spectral_partitions)
    else:
        raise Exception('Cannot process input')

    return output


def get_spectral_partitions_from_observations(
        spectra: Union[Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None) -> Union[
    Dict[str, List[List[WaveSpectrum2D]]], List[List[WaveSpectrum2D]]]:
    #

    if isinstance(spectra, dict):
        output = {}
        number = len(spectra)
        index = 0
        for key, item in spectra.items():
            index += 1
            if item is None:
                continue
            logger.info( f'{index:05d} out of {number:05d} spotter: {key}' )
            output[key] = partition_observations_spectra(item,
                                                         minimum_duration,
                                                         config,indent='    ')
    elif isinstance(spectra, list):
        output = partition_observations_spectra(spectra,
                                                minimum_duration,
                                                config)
    else:
        raise Exception('Cannot process input')

    return output


def partition_observations_spectra(spectra: List[WaveSpectrum1D],
                                   minimum_duration: timedelta,
                                   config=None, indent='') -> List[
    List[WaveSpectrum2D]]:
    if config:
        for key in config:
            assert key in default_config, f"{key} is not a valid conficuration entry"

        config = default_config | config
    else:
        config = default_config

    # Step 1: Pre-Processing
    # :
    # Prior to constructing spectra - smoothing the wave field can help create
    # more stable results.

    if config['smoothInTime']:
        logger.info( ' - Smoothing in time')
        spectra = spectrum1D_time_filter(spectra)

    # Step 2: Create 2D wavefields from 1D spectra using a spectral estimator
    spectra2D = []
    logger.info( indent + ' - Create 2D Spectra')
    for index, spectrum in enumerate(spectra):
        logger.debug( indent + f'\t {index:05d} out of {len(spectra)}')
        spectra2D.append(
            spec2d_from_spec1d(
                spectrum,
                method=config['estimator']['method'],
                number_of_directions=config['estimator']['numberOfDirections'],
                frequency_smoothing=config['estimator']['frequencySmoothing'],
                smoothing_lengthscale=config['estimator'][
                    'smoothingLengthscale']
            )
        )
    return partition_spectra(spectra2D, minimum_duration, config=config['partition_spectra'],indent=indent)





from datetime import datetime, timezone
from typing import List,Tuple,Dict,Union
from .open_remote import open_remote_forecast

def extract_along_spotter_tracks():
    pass


def extract_forecast_along_tracks( init_time ):
    pass
    # dataset = open_remote_forecast(
    #     variable=variables,
    # )


import s3fs
import h5netcdf
import numpy
from typing import List, Tuple, Union, Dict, TypedDict
from multiprocessing.pool import Pool
from .grid import interpolation_weights, interp_latitude_longitude
from itertools import repeat
from datetime import datetime
import tqdm
from roguewave.wavespectra.wavespectrum import WaveSpectrum
from roguewave.metoceandata import WaveBulkData
from pandas import DataFrame

BLOCKSIZE = 1024


def get_attr_netcdf(attrs, name, default):
    # Not all netcdf variable encode all the attributes we are looking for.
    # Here we set default values in case they are missing
    try:
        return attrs[name]
    except:
        return default


class Cluster(TypedDict):
    aws_key: str
    valid_time: datetime
    variable: str
    points: Dict[str, Tuple[float, float]]


def worker(arg) -> Dict[str, List[Union[datetime, float]]]:
    cluster = arg[0]  # type:Cluster
    latitudes = arg[1]
    longitudes = arg[2]
    variable = cluster['variable']
    filesystem = s3fs.S3FileSystem()
    out = {}

    for key, observation_coordinates in cluster['points'].items():
        if (not numpy.isnan(observation_coordinates[0])) and (
                not (numpy.isnan(observation_coordinates[1]))):
            break
    else:
        # No valid points
        out = {}
        for key in cluster['points']:
            out[key] = (cluster['valid_time'], numpy.nan)
        return out

    with filesystem.open(cluster['aws_key'], 'rb', cache_type='bytes',
                         blocksize=BLOCKSIZE) as file:
        with h5netcdf.File(file, mode='r') as ds:
            scale_factor = get_attr_netcdf(ds.variables[variable].attrs,
                                           'scale_factor', 1)
            FillValue = get_attr_netcdf(ds.variables[variable].attrs,
                                        '_FillValue', numpy.nan)
            missing_value = get_attr_netcdf(ds.variables[variable].attrs,
                                            '_missing_value', numpy.nan)
            add_offset = get_attr_netcdf(ds.variables[variable].attrs,
                                         'add_offset', 0)

            for key, observation_coordinates in cluster['points'].items():
                sum_weight = 0
                out[key] = [cluster['valid_time'], 0]

                if numpy.isnan(observation_coordinates[0]) or numpy.isnan(
                        observation_coordinates[1]):
                    out[key][1] = numpy.nan
                    continue

                weights_and_indices = interpolation_weights(
                    observation_coordinates[0], observation_coordinates[1],
                    latitudes, longitudes)

                for point in weights_and_indices:
                    ilat, ilon = point['lat_lon_indices']
                    value = ds[variable][..., ilat, ilon]
                    if isinstance(value,numpy.ndarray):
                        value = value[0]
                    # If missing/fill continue
                    if value == FillValue or value == missing_value:
                        continue

                    sum_weight += point['weight']
                    out[key][1] = out[key][1] + point['weight'] * value

                if sum_weight == 0.0:
                    out[key][1] = numpy.nan
                else:
                    out[key][1] = out[key][1] / sum_weight
                out[key][1] = out[key][1] * scale_factor + add_offset
    return out


class Track(TypedDict):
    timestamps: List[datetime]
    latitudes: numpy.ndarray
    longitudes: numpy.ndarray


def extract_at_points(
        aws_keys,
        valid_times,
        points_latlons: List[Tuple[float, float]],
        variable_name_in_netcdf,
        parallell=False,
        number_of_workers=10) -> numpy.ndarray:
    # to do the extraction, we convert to a list of the input type the
    # extraction function likes. This is just reorganizing data.
    clusters = []
    for index in range(0, len(aws_keys)):
        #
        aws_key = aws_keys[index]
        valid_time = valid_times[index]
        points = {}
        for index, point in enumerate(points_latlons):
            points[str(index)] = point

        clusters.append(
            Cluster(aws_key=aws_key, valid_time=valid_time,
                    variable=variable_name_in_netcdf,
                    points=points)
        )

    # Call the extraction function and return result.
    _data = extract_clusters(clusters, parallell=parallell,
                             number_of_workers=number_of_workers)

    number_of_points = len(points_latlons)
    number_of_times = len(valid_times)
    data = numpy.empty((number_of_points, number_of_times))

    for key, item in _data.items():
        index = int(key)
        data[index, :] = item['data'][:]

    return data


def extract_along_spotter_tracks(
        aws_keys:List[str],
        valid_times:List[datetime],
        spotter_tracks: Dict[str, Union[List[Union[WaveSpectrum, WaveBulkData]],DataFrame]],
        variable_name_in_netcdf:str,
        parallell=False,
        number_of_workers=10):
    """"""

    tracks = {}
    for spotter_id, spotter in spotter_tracks.items():
        if isinstance(spotter,DataFrame):
            latitudes = spotter['latitude'].values
            longitudes = spotter['longitude'].values
            timestamps = spotter.index
        else:
            latitudes = numpy.array([x.latitude for x in spotter])
            longitudes = numpy.array([x.longitude for x in spotter])
            timestamps = [x.timestamp for x in spotter]
        tracks[spotter_id] =Track(timestamps=timestamps, latitudes=latitudes,
                            longitudes=longitudes)

    return extract_along_tracks(aws_keys, valid_times, tracks,
                                variable_name_in_netcdf, parallell,
                                number_of_workers)


def extract_along_tracks(
        aws_keys,
        valid_times,
        tracks: Dict[str, Track],
        variable_name_in_netcdf,
        parallell=False,
        number_of_workers=10) -> Dict[str, DataFrame]:
    # The tracks are a dictionary, with as key the name of the track (e.g.
    # a Spotter id) and as contents another dictionary (see Track for description)
    # that lists for each track the times, latitudes and longitudes. Since
    # these are not on the model timebase, we first interpolate onto the timebase
    # For valid_times outside the track latitudes/longitudes we set values
    # to NaN
    for key in tracks:
        # convert the observation latitude/longitude vectors to the model
        # timebase
        tracks[key]['latitudes'] = interp_latitude_longitude(
            tracks[key]['timestamps'], tracks[key]['latitudes'], valid_times)
        tracks[key]['longitudes'] = interp_latitude_longitude(
            tracks[key]['timestamps'], tracks[key]['longitudes'], valid_times)

    # to do the extraction, we need to convert the set of tracks across model
    # results to a set of clusters we can extract per model result. This is
    # for efficiency (extract multiple points from the same model time slice)
    clusters = []
    for index in range(0, len(aws_keys)):
        #
        aws_key = aws_keys[index]
        valid_time = valid_times[index]
        points = {}
        for key, item in tracks.items():
            points[key] = (item['latitudes'][index], item['longitudes'][index])

        clusters.append(
            Cluster(aws_key=aws_key, valid_time=valid_time,
                    variable=variable_name_in_netcdf,
                    points=points)
        )

    # Call the extraction function and return result.
    return extract_clusters(clusters, parallell=parallell,
                            number_of_workers=number_of_workers)


def extract_clusters(clusters: List[Cluster],
                     parallell=False,
                     number_of_workers=10) -> Dict[
    str, DataFrame]:
    filesystem = s3fs.S3FileSystem()

    # Some properties are assumed homogeneous acrros the files. We can avoid
    # rereading them by caching from the first entry
    with filesystem.open(clusters[0]['aws_key'], 'rb',
                         cache_type='bytes', block_size=BLOCKSIZE) as file:
        with h5netcdf.File(file, mode='r') as dataset:
            latitude = dataset['latitude'][:]
            longitude = dataset['longitude'][:]

    # Create iterable that serves as input to the worker
    work = zip(clusters, repeat(latitude), repeat(longitude))
    if parallell:
        with Pool(processes=number_of_workers) as pool:
            data = list(
                tqdm.tqdm(pool.imap(worker, work),
                          total=len(clusters)))
    else:
        data = list(tqdm.tqdm(map(worker, work)))

    # Transform to a dictionary, with point names as keys and all the relevant
    # values as items
    out = {}
    for index, point_data in enumerate(data):
        for key, value in point_data.items():
            key: str
            if key not in out:
                out[key] = {'valid_times': [], 'data': []}

            out[key]['valid_times'].append(value[0])
            out[key]['data'].append(value[1])

    _out = {}
    for key, value in out.items():
        _out[key] = DataFrame(index=value['valid_times'])
        _out[key][clusters[0]['variable']] = numpy.squeeze(numpy.array(value['data']))
    return _out



from .keygeneration import generate_lead_keys, \
    generate_forecast_keys, generate_analysis_keys, \
    generate_evaluation_time_keys

from roguewave.modeldata.modelinformation import  \
    list_available_variables, available_models

from .pointdata import extract_along_spotter_tracks



"""
Contents: Simple file caching routines that automatically cache remote files
          locally for use.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Classes:
- `FileCache`, main class implementing the Caching structure. Should not
   directly be invoked. Instead, fetching/cache creation is controlled by a
   set of function defined below

Functions:

"""
import hashlib
import os
from _warnings import warn
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Union, List, Tuple, Callable
from tqdm import tqdm
from roguewave import logger
from .remote_resources import RemoteResourceS3, \
    RemoteResourceHTTPS, RemoteResource, RemoteResourceLocal
from .exceptions import _RemoteResourceUriNotFound

TEMPORARY_DIRECTORY = '~/temporary_roguewave_files/filecache/'
CACHE_SIZE_GB = 5
MAXIMUM_NUMBER_OF_WORKERS = 10
KILOBYTE = 1000
MEGABYTE = 1000 * KILOBYTE
GIGABYTE = 1000 * MEGABYTE


class FileCache:
    """
    Simple file caching class that when given an URI locally stores the
    file in the cache directory and returns the path to the file. The file
    remains in storage until the cache directory exceeds a prescribed size,
    in which case files with oldest access/modified dates get deleted first
    until everything fits in the cache again. Any time a file is accessed it's
    modified date gets updated so that often used files automaticall remain in
    cache.

    The files are stored locally in the directory specified on class
    initialization, as:

        [path]/CACHE_PREFIX + md5_hash_of_URI + CACHE_POSTFIX

    The pre- and post fix are added so we have an easy pattern to distinguish
    cache files from other files.

    Methods
      * __getitem__(keys) : accept a simgle uri_key or a list of URI's and
        returns filepaths to local copies thereof. You would typically use the
            cache[keys] notation instead of the dunder method.
      * purge() clear all contents of the cache (destructive, deletes all local
        files).

    Usage:

        cache = FileCache()
        list_of_local_file_names = cache[ [list of URI's ] ]

    # do stuff with file
    ...
    """

    CACHE_FILE_PREFIX = 'cachefile_'
    CACHE_FILE_POSTFIX = '_cachefile'

    def __init__(self,
                 path: str = TEMPORARY_DIRECTORY,
                 size_GB: Union[float, int] = CACHE_SIZE_GB,
                 do_cache_eviction_on_startup: bool = False,
                 resources:List[RemoteResource]=None,
                 parallel=True,
                 allow_for_missing_files=False,
                 post_process_function: Callable[[str],None] = None,
                 validate_function: Callable[[str], None] = None,
                 ):
        """
        Initialize Cache
        :param path: path to store cache. If path does not exist it will be
            created.
        :param size_GB: Maximum size of the cache in GiB. If cache exceeds
            the size, then files with oldest access/modified dates get deleted
            until everthing fits in the cache again. Fractional values (floats)
            are allowed.
        :param do_cache_eviction_on_startup: whether we ensure the cache size
            conforms to the given size on startup. If set to true, a cache
            directory that exceeds the maximum size will be reduced to max
            size. Set to False by default in which case an error occurs. The
            latter to prevent eroneously evicting files from a cache that was
            previously created on purpose with a larger size that the limit.


        """

        self.path = os.path.expanduser(path)
        self.max_size = int(size_GB * GIGABYTE)
        self.parallel = parallel
        self.allow_for_missing_files = allow_for_missing_files

        # create the path if it does not exist
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        # Some counters to keep track of total cache misses, hits and
        # evictions. No downstream use right now/
        self._cache_misses = 0
        self._cache_hits = 0
        self._cache_evictions = 0

        # initialize the cache.
        self._entries = {}  # the key/value pair cache
        self._initialize_cache(do_cache_eviction_on_startup)

        # Post processing and validation functions
        self.post_process_function = post_process_function
        self.validate_function = validate_function

        # download resources
        if resources is None:
            self.resources = [RemoteResourceS3(),
                              RemoteResourceHTTPS(),
                              RemoteResourceLocal()]
        else:
            self.resources = resources

    def _post_process(self,filepath:str):
        """
        Call custom post processing function that gets called after first
        download- if set.
        :param filepaths: list of filepaths
        :return: none
        """
        if self.post_process_function is None:
            return
        else:
            self.post_process_function(filepath)

    def _validate(self,filepath:str)->bool:
        """
        Call custom post processing function that gets called after first
        download- if set.
        :param filepaths: list of filepaths
        :return: none
        """
        if self.validate_function is None:
            return True
        else:
            return self.validate_function(filepath)

    def _cache_file_name(self, uri: str) -> str:
        """
        Return the filename that corresponds to the given uri. We construct
        the file name using a simple md5 hash of the uri string prefixed
        with a cache file prefix. THe later is introduced to seperate cache
        files in a path from user files (and avoid including/deleting those).

        :param uri: valis uri"
        :return: valid cache file
        """
        return self.CACHE_FILE_PREFIX + _hashname(
            uri) + self.CACHE_FILE_POSTFIX

    def _cache_file_path(self, uri: str) -> str:
        """
        Construct the path where the given uri is stored locally.

        :param uri: valis uri of form "bucket/key"
        :return: valid cache file
        """
        return os.path.join(self.path, self._cache_file_name(uri))

    def _get_cache_files(self) -> List[str]:
        """
        Find all files that are currently a member of the cache.
        :return:
        """
        _cache_files = []
        for path, dirs, files in os.walk(self.path):
            # Return all files that are "cache" objects. This is a safety if
            # other user files are present, so that these don't accidentally
            # evicted from the cache (aka deleted).
            return [file for file in files if
                    file.startswith(self.CACHE_FILE_PREFIX) and
                    file.endswith(self.CACHE_FILE_POSTFIX)]
        else:
            return []

    def _initialize_cache(self, do_cache_eviction_on_startup: bool) -> None:
        """
        Initialize the file cache. Look on disk for files in the cache path
        that have the required prefix and load these into the cache. Once
        loaded, we do a check whether or not the cache is full and if we need
        to remove files.
        :param do_cache_eviction_on_startup: see description under __init__
        :return:
        """
        self._entries = {}
        for file in self._get_cache_files():
            filepath = os.path.join(self.path, file)
            self._entries[file] = filepath

        # See if cache is still < size
        if do_cache_eviction_on_startup:
            self._cache_eviction()
        else:
            if self._size() > self.max_size:
                raise ValueError('The cache currently existing on disk '
                                 'exceeds the maximum cache size of the '
                                 'current cache.\n Either increase the cache '
                                 'size of the object or allow eviction of '
                                 'files on startup.')

    def in_cache(self, uris) -> List[bool]:
        # make sure input is a list
        if isinstance(uris, str):
            uris = [uris]

        # Create the hashes from the URI's
        hashes = [self._cache_file_name(uri) for uri in uris]
        return [self._is_in_cache(_hash) for _hash in hashes]

    def _is_in_cache(self, _hash: str) -> bool:
        """
        Check if a _hash is in the cache
        :param _hash: hash to check
        :return: True if in Cache, False if not.
        """
        cache_hit = _hash in self._entries
        return cache_hit

    def _add_to_cache(self, _hash: str, filepath: str) -> None:
        # add entry to the cache.
        self._entries[_hash] = filepath

    def remove(self, uri: str) -> None:
        """
        Remove an entry from the cache
        :param uri: uri
        :return: None
        """
        if not self.in_cache(uri):
            raise ValueError(f'Key {uri} not in Cache')

        _hash = self._cache_file_name(uri)
        return self._remove_item_from_cache(_hash)

    def _remove_item_from_cache(self, _hash: str) -> None:
        """
        Remove a hash key from the cache. Here it is assumed that the _hash is
        a valid entry. We do allow for non existance of corresponding files as
        the cache can get out of sync if something external deleted the file.
        Since the endstate is valid (no entry in cache, no entry on disk) this
        is considered OK.

        :param _hash: hash key
        :return: None
        """

        assert _hash in self._entries

        file_to_delete = self._entries.pop(_hash)

        if os.path.exists(file_to_delete):
            logger.debug(f' - removing {_hash}')

            # And delete file.
            os.remove(file_to_delete)
        else:
            logger.debug(f' - file {_hash} did not exist on disk')

        return None

    def _get_from_cache(self, _hash: str) -> str:
        """
        Get entry from cache and touch the file to indicate it has been used
        recently (last to be evicted)
        :param _hash: file_hash corresponding to uri
        :return: file path
        """
        filepath = self._entries[_hash]

        if not os.path.exists(filepath):
            raise FileNotFoundError('The filepath in the cache log does not'
                                    'exist on disk.')

        # Touch the file to indicate we recently used it.
        Path(filepath).touch()

        return filepath

    def __len__(self) -> int:
        """
        :return: Number of entries in the cache.
        """
        return len(self._entries)

    def __getitem__(self, uris: Union[List, str]) -> List[str]:

        # make sure input is a list
        if isinstance(uris, str):
            uris = [uris]

        # Create the hashes from the URI's
        hashes = [self._cache_file_name(uri) for uri in uris]

        # collect key, full local path and hash for any hashes that are not
        # in the local cache
        cache_misses = []
        for _hash, key in zip(hashes, uris):
            to_append = (key, self._cache_file_path(key), _hash)
            if self._is_in_cache(_hash):
                if not self._validate( self._entries[_hash] ):
                    self._remove_item_from_cache(_hash)
                    cache_misses.append(to_append)
            else:
                cache_misses.append(to_append)

        self._cache_misses += len(cache_misses)
        self._cache_hits += len(hashes) - len(cache_misses)

        # for all URI's not in cache
        if cache_misses:
            # download all the files
            succesfully_downloaded = _download_from_resources(
                cache_misses,
                self.resources,
                parallel_download=self.parallel,
                allow_for_missing_files=self.allow_for_missing_files)

            # For all downloaded files do
            for success, cache_miss in zip(
                    succesfully_downloaded, cache_misses):
                if success:
                    # If succesfull, add to cache.
                    self._post_process(cache_miss[1])
                    self._add_to_cache(cache_miss[2], cache_miss[1])
                else:
                    # If not succesful, remove from keys to return
                    # Todo some logging or erroring?
                    hashes.pop(hashes.index(cache_miss[2]))

        # Get the filepaths to return
        filepaths = [self._get_from_cache(_hash) for _hash in hashes]

        size_of_requested_data = _get_total_size_of_files_in_bytes(filepaths)
        if size_of_requested_data > self.max_size:
            warning = f'The requested data does not fit into the cache.' \
                      f'To avoid issues the cache is enlarged to ensure' \
                      f'the current set of files fits in the cache. \n' \
                      f'old size: {self.max_size} bytes; ' \
                      f'new size {size_of_requested_data + MEGABYTE}'
            warn(warning)
            logger.warning(warning)
            self.max_size = size_of_requested_data + MEGABYTE

        # See if we need to do any cache eviction because the cache has become
        # to big.
        self._cache_eviction()
        return filepaths

    def _cache_eviction(self) -> bool:
        """
        Simple cache eviction policy. If the cache exceeds the maximum size
        remove data from the cache based on whichever file was interacted with
        the longest time ago. Evict files until we are below the acceptable
        cache size.

        :return: True if eviction occured, False otherwise.
        """

        # check if we exceed the size, if not return
        if not self._size() > self.max_size:
            return False

        # Get access/modified times for all the files in cache
        modified = []
        for _hash, fp in self._entries.items():
            # From my brief reading, access time is not always reliable,
            # hence I use whatever the latest time set is for modified or
            # access time as an indicator of when we last interacted with
            # the file.
            access_time = os.path.getatime(fp)
            modified_time = os.path.getmtime(fp)

            # pick whichever is most recent.
            time_to_check = access_time if access_time > modified_time \
                else modified_time
            modified.append((time_to_check, _hash))

        # Sort files in reversed chronological order.
        files_in_cache = [x[1] for x in
                          sorted(modified, key=lambda x: x[0], reverse=True)]

        # Delete files one by one as long as the cache_size exceeds the max
        # size.
        while (_size := self._size()) > self.max_size:
            self._cache_evictions += 1
            logger.debug(
                f'Cache exceeds limits: {_size} bytes, max size: '
                f'{self.max_size} bytes')

            # Get the hash and path of the oldest file and remove
            self._remove_item_from_cache(files_in_cache.pop())

        return True

    def _size(self) -> int:
        """
        Return size on disk of the cache in bytes.
        :return: cache size in bytes.
        """
        return _get_total_size_of_files_in_bytes(
            list(self._entries.values()), self.path)

    def purge(self) -> None:
        """
        Delete all the files in the cache.
        :return: None
        """
        logger.debug(f'Purging cache')
        keys = list(self._entries.keys())
        for key in keys:
            filepath = self._entries.pop(key)
            logger.debug(f' - deleting {filepath}')
            os.remove(filepath)
        logger.debug(f'Purging cache done')


def _download_from_resources(key_and_filenames: List[Tuple],
                             resources: List[RemoteResource],
                             parallel_download=False,
                             allow_for_missing_files=False) -> List[bool]:
    """
    Wrapper function to download multiple uris from the resource(s).

    :param key_and_filenames: List containing (uri,filename)
    :param parallel_download: If true, downloading is performed in parallel.
    :return: List of boolean indicating if the download was a success.
    """

    # construct the arguments to be used for parallel downloading of files.
    # Specifically, we need to match the right resource for downloading to the
    # right URI.
    args = []
    for key_and_filename in key_and_filenames:
        # Loop over all resources until we find one that can interpret the URI
        # (this is pretty naive approach and should probably be refactored to
        #  some direct mapping if the number of resources ever gets very long)
        for resource in resources:
            # For each resource check if the resource can interpret the URI
            if resource.valid_uri(key_and_filename[0]):
                # If so, get the download function, and other arguments and
                # break
                args.append(
                    (resource.download(), *key_and_filename[0:2],
                     allow_for_missing_files)
                )
                break
        else:
            # If we didn't break the loop no valid resource was found, raise
            # error
            raise ValueError(f'No resource available for URI: '
                             f'{key_and_filename[0]}')

    # Download the requested objects.
    if parallel_download:
        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
            output = list(tqdm(pool.imap(_worker, args), total=len(args)))
    else:
        output = list(tqdm(map(_worker, args),total=len(args)))

    return output


def _get_total_size_of_files_in_bytes(filenames: List[str], path=None) -> int:
    """
    Simple function to calculate the size of a list of files on disk.
    :param filenames: list of filenames or filepaths
    :param path: if filenames are provided, this lists the path, otherwise set
        to None

    :return: Total size in bytes
    """
    size = 0
    for filename in filenames:
        if path is None:
            filepath = filename
        else:
            filepath = os.path.join(path, filename)
        size += os.path.getsize(filepath)
    return size


def _hashname(string: str) -> str:
    """
    Returns a md5 hash of a given string.
    :param string: input string
    :return: hexdigest of md5 hash.
    """
    return hashlib.md5(string.encode(), usedforsecurity=False).hexdigest()


def _worker(args) -> bool:
    download_func = args[0]
    uri = args[1]
    filepath = args[2]
    allow_for_missing_files = args[3]
    try:
        download_func(uri, filepath)
        return True
    except _RemoteResourceUriNotFound as e:
        if allow_for_missing_files:
            warning = f'Uri not retrieved: {str(e)}'
            warn(warning)
            logger.warning(warning)
        else:
            raise e
        return False


# Import
# =============================================================================
from datetime import datetime, timedelta

from .modelinformation import _get_resource_specification, \
    _generate_aws_key
from .timebase import timebase_forecast, \
    timebase_lead, timebase_evaluation
from typing import List, Tuple


# Main Public Functions
# =============================================================================
def generate_forecast_keys(
        variable: str,
        init_time: datetime,
        duration: timedelta,
        model_name: str) -> List[str]:
    """
    Get the AWS keys associated with a forecast for a given
    model.

    :param variable: name of the variable of interest
    :param init_time: init time of the forecast of interest
    :param duration: maximum lead time of interest
    :param model_name: model name
    :return: List of valid aws_keys.
    """
    aws_key_layout = _get_resource_specification(model_name)
    time_vector = timebase_forecast(
        init_time=init_time,
        duration=duration,
        time_configuration=aws_key_layout.model_time_configuration
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# ------------------------------------------------------------------------------


def generate_analysis_keys(
        variable:str,
        start_time: datetime,
        end_time: datetime,
        model_name: str,
) -> List[str]:
    """
    Get the AWS keys associated with the analysis results for
    a given model.

    :param variable: name of the variable of interest
    :param start_time: Start of time interval of interest
    :param end_time: End of time interval of interest
    :param model_name: model name
    :return: List of the aws_keys
    """
    aws_key_layout = _get_resource_specification(model_name)
    time_vector = timebase_lead(
        start_time,
        end_time,
        timedelta(hours=0),
        time_configuration=aws_key_layout.model_time_configuration,
        exact=False
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# -----------------------------------------------------------------------------


def generate_lead_keys(
        variable: str,
        start_time: datetime, end_time: datetime,
        lead_time: timedelta, model_name: str,
        exact=False) -> List[str]:
    """
    Get the AWS keys associated with a given lead time for a
    given model.

    :param variable: name of the variable of interest
    :param start_time: Start of time interval of interest
    :param end_time: End of time interval of interest
    :param lead_time: Lead time of interest
    :param model_name: model name
    :param exact: If true, generate keys exactly at given lead time. Otherwise
    generate keys as close as possible.
    :return: List of the aws_keys.
    """
    aws_key_layout = _get_resource_specification(model_name)
    time_vector = timebase_lead(
        start_time=start_time,
        end_time=end_time,
        lead_time=lead_time,
        time_configuration=aws_key_layout.model_time_configuration,
        exact=exact
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# -----------------------------------------------------------------------------


def generate_evaluation_time_keys(
        variable: str,
        evaluation_time: datetime,
        model_name: str,
        maximum_lead_time: timedelta = None
) -> List[str]:
    """
    Get the AWS keys associated with a given evaluation time for a
    given model.

    :param variable: name of the variable of interest
    :param evaluation_time: evaluation time of interest
    :param model_name: model name
    :param maximum_lead_time: maximum lead time of interest
    :return: list of aws keys
    """
    aws_key_layout = _get_resource_specification(model_name)
    if maximum_lead_time is None:
        maximum_lead_time = aws_key_layout.model_time_configuration.duration

    time_vector = timebase_evaluation(
        evaluation_time=evaluation_time,
        maximum_lead_time=maximum_lead_time,
        time_configuration=aws_key_layout.model_time_configuration
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# -----------------------------------------------------------------------------


# Internal module functions
# =============================================================================
def _generate_keys(time_vector, variable, aws_key_layout) -> List[str]:
    aws_keys = []
    for time in time_vector:
        aws_keys.append(
            _generate_aws_key(
                variable=variable,
                init_time=time[0],
                forecast_hour=time[1],
                model=aws_key_layout)
        )
    return aws_keys

